// Copyright (C) 2018  Nexedi SA and Contributors.
//                     Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

package main
// ZBlk* + ZBigFile loading


import (
	"context"
	"fmt"
	"reflect"
	"sort"
	"sync"

	"golang.org/x/sync/errgroup"

	"lab.nexedi.com/kirr/go123/mem"
	"lab.nexedi.com/kirr/neo/go/zodb"
	"lab.nexedi.com/kirr/neo/go/zodb/btree"
	pickle "github.com/kisielk/og-rek"
)

// module of Wendelin ZODB py objects
const zwendelin = "wendelin.bigfile.file_zodb"

// ---- ZBlk0 ----

// ZBlk0 mimics ZBlk0 from python.
type ZBlk0 struct {
	zodb.Persistent

	blkdata string
}

type zBlk0State ZBlk0 // hide state methods from public API

func (zb *zBlk0State) DropState() {
	zb.blkdata = ""
}

func (zb *zBlk0State) PySetState(pystate interface{}) error {
	blkdata, ok := pystate.(string)
	if !ok {
		return fmt.Errorf("expect str; got %T", pystate)
	}

	zb.blkdata = blkdata
	return nil
}

func (zb *ZBlk0) LoadBlkData(ctx context.Context) ([]byte, error) {
	// XXX err ctx

	err := zb.PActivate(ctx)
	if err != nil {
		return nil, err
	}
	defer zb.PDeactivate()

	return mem.Bytes(zb.blkdata), nil
}

// ---- ZBlk1 ---

// ZData mimics ZData from python.
type ZData struct {
	zodb.Persistent

	data string
}

type zDataState ZData // hide state methods from public API

func (zd *zDataState) DropState() {
	zd.data = ""
}

func (zd *zDataState) PySetState(pystate interface{}) error {
	data, ok := pystate.(string)
	if !ok {
		return fmt.Errorf("expect str; got %T", pystate)
	}

	zd.data = data
	return nil
}

// ZBlk1 mimics ZBlk1 from python.
type ZBlk1 struct {
	zodb.Persistent

	chunktab *btree.IOBTree // {} offset -> ZData(chunk)
}

type zBlk1State ZBlk1 // hide state methods from public API

func (zb *zBlk1State) DropState() {
	zb.chunktab = nil
}

func (zb *zBlk1State) PySetState(pystate interface{}) error {
	chunktab, ok := pystate.(*btree.IOBTree)
	if !ok {
		//return fmt.Errorf("expect IOBTree; got %T", pystate)
		return fmt.Errorf("expect IOBTree; got %#v", pystate)
	}

	zb.chunktab = chunktab
	return nil
}

func (zb *ZBlk1) LoadBlkData(ctx context.Context) ([]byte, error) {
	// XXX errctx

	err := zb.PActivate(ctx)
	if err != nil {
		return nil, err
	}
	defer zb.PDeactivate()

	// // empty .chunktab -> ø
	// ev := zb.chunktab.Entryv()
	// if len(ev) == 0 {
	// 	return nil, nil
	// }

	// get to all ZData objects; activate them and build
	//
	//	{} offset -> ZData
	//
	// with all ZData being live.
	var mu sync.Mutex
	chunktab := make(map[int64]*ZData)

	// on return deactivate all loaded ZData objects in chunktab
	defer func() {
		for _, zd := range chunktab {
			zd.PDeactivate()
		}
	}()


	wg, ctx := errgroup.WithContext(ctx)

	// loadZData loads 1 ZData object into chunktab and leaves it activated.
	loadZData := func(offset int64, zd *ZData) error {
		err := zd.PActivate(ctx)
		if err != nil {
			return err
		}
		// no PDeactivate, zd remains live

		fmt.Printf("@%d -> zdata #%s (%d)\n", offset, zd.POid(), len(zd.data))

		mu.Lock()
		defer mu.Unlock()

		// check dup keys
		if _, already := chunktab[offset]; already {
			panic(fmt.Sprintf("already %v", offset))	// XXX
		}
		chunktab[offset] = zd
		return nil
	}

	// loadBucket loads all ZData objects from leaf BTree bucket.
	loadBucket := func(b *btree.IOBucket) error {
		err := b.PActivate(ctx)
		if err != nil {
			return err
		}
		defer b.PDeactivate()

		// XXX go through all bucket key/v -> chunktab
		// XXX off	 < 0		!ok
		// XXX off + len > blksize	!ok

		fmt.Printf("\nbucket: %v\n", b.Entryv())
		for _, e := range b.Entryv() {
			zd, ok := e.Value().(*ZData)
			if !ok {
				return fmt.Errorf("!ZData (%T)", e.Value())	// XXX
			}

			wg.Go(func() error {
				return loadZData(int64(e.Key()), zd)	// XXX cast
			})
		}

		return nil
	}


	// loadBTree spawns loading of all BTree children.
	var loadBTree func(t *btree.IOBTree) error
	loadBTree = func(t *btree.IOBTree) error {
		err := t.PActivate(ctx)
		if err != nil {
			return err
		}
		defer t.PDeactivate()

		fmt.Printf("\nbtree: %v\n", t.Entryv())

		for _, e := range t.Entryv() {
			switch child := e.Child().(type) {
			case *btree.IOBTree:
				wg.Go(func() error {
					return loadBTree(child)
				})

			case *btree.IOBucket:
				wg.Go(func() error {
					return loadBucket(child)
				})

			default:
				panic(0)	// XXX
			}
		}

		return nil
	}

	wg.Go(func() error {
		return loadBTree(zb.chunktab)
	})

	err = wg.Wait()
	if err != nil {
		return nil, err // XXX err ctx
	}

	// empty .chunktab -> ø
	if len(chunktab) == 0 {
		return nil, nil
	}

	// glue all chunks from chunktab
	offv := make([]int64, 0, len(chunktab)) // ↑
	for off := range(chunktab) {
		offv = append(offv, off)
	}
	sort.Slice(offv, func(i, j int) bool {
		return offv[i] < offv[j]
	})

	fmt.Printf("#chunktab: %d\n", len(chunktab))
	fmt.Printf("offv: %v\n", offv)


        // find out whole blk len via inspecting tail chunk
	tailStart := offv[len(offv)-1]
	tailChunk := chunktab[tailStart]
	blklen    := tailStart + int64(len(tailChunk.data))

        // whole buffer initialized as 0 + tail_chunk
	blkdata := make([]byte, blklen)
	copy(blkdata[tailStart:], tailChunk.data)

        // go through all chunks besides tail and extract them
	stop := int64(0)
	for _, start := range offv[:len(offv)-1] {
		chunk := chunktab[start]
		if !(start >= stop) {	// verify chunks don't overlap
			panic("!(start >= stop)")	// XXX
		}
		if !(start + int64(len(chunk.data)) <= int64(len(blkdata))) {
			panic("blkdata overrun")	// XXX
		}
		stop = start + int64(len(chunk.data))
		copy(blkdata[start:], chunk.data)
	}

	return blkdata, nil
}


// ----------------------------------------

// ZBigFile mimics ZBigFile from python.
type ZBigFile struct {
	zodb.Persistent

	blksize	int64
	blktab  *btree.LOBTree	// {}  blk -> ZBlk*(blkdata)
}

type zBigFileState ZBigFile // hide state methods from public API


// DropState implements Stateful.
func (bf *zBigFileState) DropState() {
	bf.blksize = -1
	bf.blktab  = nil
}

// PySetState implements PyStateful.
func (bf *zBigFileState) PySetState(pystate interface{}) (err error) {
	// ZBigFile
	//	.blksize	xint
	//	.blktab		LOBtree{}  blk -> ZBlk*(blkdata)
	//
	// state: (.blksize, .blktab)

	t, ok := pystate.(pickle.Tuple)
	if !ok {
		return fmt.Errorf("expect [2](); got %T", pystate)
	}
	if len(t) != 2 {
		return fmt.Errorf("expect [2](); got [%d]()", len(t))
	}

	// blksize, ok := pickletools.Xint64(t[0])	// XXX reenable (pickletools)
	if !ok {
		return fmt.Errorf("blksize: expect integer; got %T", t[0])
	}

	blktab, ok := t[1].(*btree.LOBTree)
	if !ok {
		return fmt.Errorf("blktab: expect LOBTree; got %T", t[1])
	}

	bf.blksize = 0		// XXX blksize
	bf.blktab  = blktab
	return nil
}


// ----------------------------------------

func init() {
	t := reflect.TypeOf
	zodb.RegisterClass(zwendelin + ".ZBlk",     t(ZBlk0{}),    t(zBlk0State{}))
	zodb.RegisterClass(zwendelin + ".ZBlk0",    t(ZBlk0{}),    t(zBlk0State{}))
	zodb.RegisterClass(zwendelin + ".ZBlk1",    t(ZBlk1{}),    t(zBlk1State{}))
	zodb.RegisterClass(zwendelin + ".ZData",    t(ZData{}),    t(zDataState{}))
	zodb.RegisterClass(zwendelin + ".ZBigFile", t(ZBigFile{}), t(zBigFileState{}))
}
